# Some common procedures used for accessing the control system's MPX subsystem
# This version uses the new control system EPICS SOAP server access method
# Modifying these procs may seem an odd approach and may not be effiecient
# but it should mean few, if any, changes elsewhere
# The method is neither extensible nor flexible. It expects the DAQ system
# to control FaradayCup (meaning the last Faraday Cup - No.4, B-FC-301)
# and the Buzzer only, via write-mpx-bits & write-mpx-reg respectively

# server name should not be built in
# need to change this to pick it up from somewhere (experiment?)
set epicsServer "nndhcp069"

# The existing system will write 0/1 to turn the buzzer Off/On
# The existing system will write 0/1 to move the cup In/Out
# The DAQ code uses the names Buzzer.switch & FaradayCup.control

set EPICS(Buzzer.switch) [list MEIS-CON-A-1:CON:ON {Off On}]
set EPICS(FaradayCup.control) [list MEIS-B-FC-301:SOL:CON:IN {In Out}]

set EPICSRegisters {}

# Two epics access routines taken from MEISControls - meisc.common
set epicsInitialised 0

proc epicsInitServer {} {
    global epicsServer
    set server [inform sys inquire epicsServer]
    if {[llength $server] != 1} {
        set server $epicsServer
    }
    epicsNewServer $server
}

proc epicsNewServer {newServer} {
    global epicsServer
    set epicsServer $newServer
    epics_initialise $epicsServer 1
    inform sys set_Options epicsServer $epicsServer
    inform sys update_Options
    define-mpx-controls
}

proc epics_initialise {{server localhost} args} {
    global EPICSAccessServerHost EPICSServer
    global EPICSPVNames
    global env epicsInitialised

    if {[llength $args] != 0 && $epicsInitialised} {return}
    set epicsInitialised 1

    set EPICSAccessServerHost $server
    package require EPICSAccessClient

    loadsource EPICSAccessClient     ;# better solution needed

    set EPICSServer $server

    if {$EPICSServer == "localhost"} {
          source [file join $env(MIDASBASE) MIDASpackages EPICSClientAccess1.0 pkgIndex.tcl]
          package require EPICSClientAccess
    }
}

proc epics_put_string {PVN String} {
    upvar #0 MIDAS_SOAPFaultCode FaultCode
    upvar #0 MIDAS_SOAPFaultString FaultString
    upvar #0 MIDAS_SOAPErrorCode ErrorCode
    upvar #0 MIDAS_SOAPErrorInfo ErrorInfo
    upvar #0 EPICS_Verbose Verbose
    global MIDAS_EPICSAlarmSeverity MIDAS_EPICSAlarmStatus

    set rc [catch {EPICSAccessClient__CreateChannel $PVN} m]
    if {$rc != 0} {
        insert-debug "CreateChannel for $PVN returned with an error\n$m\n$FaultCode\n$FaultString\n$ErrorCode\n$ErrorInfo"
        return
    } else {
       if {$Verbose} {insert-debug "CreateChannel returned $m"}
    }

    set rc [catch {EPICSAccessClient__PutString $PVN $String} m]
    if {$rc != 0} {
        insert-debug "PutString for $PVN returned with an error\n$m\n$FaultCode\n$FaultString\n$ErrorCode\n$ErrorInfo"
    } else {
        if {$Verbose} {insert-debug "PutString returned $m"}
        global $PVN; set $PVN $String
        global [set PVN]#ASEVR; set [set PVN]#ASEVR [set MIDAS_EPICSAlarmSeverity]
        global [set PVN]#ASTAT; set [set PVN]#ASTAT [set MIDAS_EPICSAlarmStatus]
    }

    return $m
}

#

proc write-epics-string {r v} {
   global EPICSRegisters $r

   set f [lsearch $EPICSRegisters $r]
   if {$f == -1} {
      return -code error -errorcode 0x20004 -errorinfo "Register not known"
   }
   set PV_Record [first [set $r]]
   set Strings   [second [set $r]]
   set v [expr $v != 0]
   epics_put_string $PV_Record [lindex $Strings $v]
}

proc define-mpx-register {mname class rname attributes} {
   global epicsServer EPICSRegisters EPICS $mname.$rname

   epics_initialise $epicsServer

   if {![info exists EPICS($mname.$rname)]} {
      return -code error -errorcode 0x2000b -errorinfo "Illegal register name"
   }
   set $mname.$rname $EPICS($mname.$rname)
   lappend EPICSRegisters $mname.$rname
}

proc write-mpx-reg {r v} {
# v == 0 => Buzzer Off, !0 => Buzzer On
   if {![string equal $r "Buzzer.switch"]} {
      return -code error -errorcode 0x20004 -errorinfo "Register not known"
   }
   write-epics-string $r $v
}

proc write-mpx-bits {r m sc} {
# m was a mask defining which bits should be selectively modified
# sc specifies whether they should be cleared (sc==0) or set (sc!=0)
# sc == 0 => Cup In, !0 => Cup out
   if {![string equal $r "FaradayCup.control"]} {
      return -code error -errorcode 0x20004 -errorinfo "Register not known"
   }
   write-epics-string $r $sc
}

proc query-mpx-names {p} {
   global EPICSRegisters
   return [lindex $EPICSRegisters [lsearch $EPICSRegisters $p]]
}

epicsInitServer
